@@ -1,10 +1,12 @@
 use strict;
 use warnings;
-package Test::Warnings; # git description: v0.020-10-g58c8076
+package Test::Warnings; # git description: v0.027-4-g76bf83a
+# vim: set ts=8 sts=4 sw=4 tw=115 et :
 # ABSTRACT: Test for warnings and the lack of them
 # KEYWORDS: testing tests warnings
-# vim: set ts=8 sts=4 sw=4 tw=78 et :
-our $VERSION = '0.021';
+
+our $VERSION = '0.028';
+
 use parent 'Exporter';
 use Test::Builder;
 
@@ -19,14 +21,20 @@ my $warnings_allowed;
 my $forbidden_warnings_found;
 my $done_testing_called;
 my $no_end_test;
+my $fail_on_warning;
 
 sub import
 {
-    # END block will check for this status
-    my @symbols = grep { $_ ne ':no_end_test' } @_;
-    $no_end_test = (@symbols != @_);
+    my $class = shift @_;
 
-    __PACKAGE__->export_to_level(1, @symbols);
+    my %names; @names{@_} = ();
+    # END block will check for this status
+    $no_end_test = exists $names{':no_end_test'};
+    # __WARN__ handler will check for this status
+    $fail_on_warning = exists $names{':fail_on_warning'};
+
+    delete @names{qw(:no_end_test :fail_on_warning)};
+    __PACKAGE__->export_to_level(1, $class, keys %names);
 }
 
 # for testing this module only!
@@ -42,28 +50,41 @@ sub _builder(;$)
     $tb = shift;
 }
 
+my $_orig_warn_handler = $SIG{__WARN__};
 $SIG{__WARN__} = sub {
-    my $msg = shift;
-
     if ($warnings_allowed)
     {
-        Test::Builder->new->note($msg);
+        Test::Builder->new->note($_[0]);
     }
     else
     {
-        if ($msg =~ /\n$/) {
-            warn $msg;
+        $forbidden_warnings_found++;
+
+        # TODO: this doesn't handle blessed coderefs... does anyone care?
+        goto &$_orig_warn_handler if $_orig_warn_handler
+            and (  (ref $_orig_warn_handler eq 'CODE')
+                or ($_orig_warn_handler ne 'DEFAULT'
+                    and $_orig_warn_handler ne 'IGNORE'
+                    and defined &$_orig_warn_handler));
+
+        if ($_[0] =~ /\n$/) {
+            warn $_[0];
         } else {
             require Carp;
-            Carp::carp($msg);
+            Carp::carp($_[0]);
         }
-
-        $forbidden_warnings_found++;
+        _builder->ok(0, 'unexpected warning') if $fail_on_warning;
     }
 };
 
-sub warnings(&)
+sub warnings(;&)
 {
+    # if someone manually does warnings->import in the same namespace this is
+    # imported into, this sub will be called.  in that case, just return the
+    # string "warnings" so it calls the correct method.
+    if (!@_) {
+        return 'warnings';
+    }
     my $code = shift;
     my @warnings;
     local $SIG{__WARN__} = sub {
@@ -146,7 +167,7 @@ Test::Warnings - Test for warnings and the lack of them
 
 =head1 VERSION
 
-version 0.021
+version 0.028
 
 =head1 SYNOPSIS
 
@@ -321,6 +342,16 @@ Imports all functions listed above
 Disables the addition of a C<had_no_warnings> test
 via C<END> or C<done_testing>
 
+=head2 C<:fail_on_warning>
+
+=for stopwords unexempted
+
+When used, fail immediately when an unexempted warning is generated (as opposed to waiting until
+L</had_no_warnings> or C<done_testing> is called).
+
+I recommend you only turn this option on when debugging a test, to see where a surprise warning is coming from,
+and rely on the end-of-tests check otherwise.
+
 =head1 CAVEATS
 
 =for stopwords smartmatch TODO irc
@@ -377,13 +408,6 @@ environment
 
 =back
 
-=head1 SUPPORT
-
-=for stopwords irc
-
-Bugs may be submitted through L<https://rt.cpan.org/Public/Dist/Display.html?Name=Test-Warnings>.
-I am also usually active on irc, as 'ether' at C<irc.perl.org>.
-
 =head1 SEE ALSO
 
 =for stopwords YANWT
@@ -404,9 +428,7 @@ L<blogs.perl.org: YANWT (Yet Another No-Warnings Tester)|http://blogs.perl.org/u
 
 =item *
 
-L<strictures> - which makes all warnings fatal in tests, hence lessening
-
-the need for special warning testing
+L<strictures> - which makes all warnings fatal in tests, hence lessening the need for special warning testing
 
 =item *
 
@@ -418,21 +440,48 @@ L<Test::Fatal>
 
 =back
 
+=head1 SUPPORT
+
+Bugs may be submitted through L<the RT bug tracker|https://rt.cpan.org/Public/Dist/Display.html?Name=Test-Warnings>
+(or L<bug-Test-Warnings@rt.cpan.org|mailto:bug-Test-Warnings@rt.cpan.org>).
+
+There is also a mailing list available for users of this distribution, at
+L<http://lists.perl.org/list/perl-qa.html>.
+
+There is also an irc channel available for users of this distribution, at
+L<C<#perl> on C<irc.perl.org>|irc://irc.perl.org/#perl-qa>.
+
+I am also usually active on irc, as 'ether' at C<irc.perl.org>.
+
 =head1 AUTHOR
 
 Karen Etheridge <ether@cpan.org>
 
-=head1 COPYRIGHT AND LICENSE
+=head1 CONTRIBUTORS
+
+=for stopwords Graham Knop A. Sinan Unur Leon Timmermans
+
+=over 4
+
+=item *
+
+Graham Knop <haarg@haarg.org>
+
+=item *
+
+A. Sinan Unur <nanis@cpan.org>
+
+=item *
+
+Leon Timmermans <fawaka@gmail.com>
+
+=back
+
+=head1 COPYRIGHT AND LICENCE
 
 This software is copyright (c) 2013 by Karen Etheridge.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
 
-=head1 CONTRIBUTOR
-
-=for stopwords A. Sinan Unur
-
-A. Sinan Unur <nanis@cpan.org>
-
 =cut
